/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: EjbBC.java ofabre$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.bc.ejb;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import javax.jbi.JBIException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.spi.InitialContextFactory;
import javax.rmi.PortableRemoteObject;
import org.ow2.petals.bc.ejb.constants.EjbVersion;
import org.ow2.petals.bc.ejb.constants.ProvidesConstants;
import org.ow2.petals.bc.ejb.su.EjbServiceUnitListener;
import org.ow2.petals.bc.ejb.xmlengine.XMLStream;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.bc.AbstractBindingComponent;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.su.BindingComponentServiceUnitManager;
import org.ow2.petals.databinding.jaxb.service.Service;





/**
 * 
 * @author ofabre & ccasadei - eBM WebSourcing
 */
public class EjbBC extends AbstractBindingComponent {

    private Map<Provides, Object> ejbPool;

    private Map<Provides, XMLStream> xmlEnginePool;

    private Map<Provides, Service> dataBindingServicePool;
    
    private Map<String, ClassLoader> suClassLoaders;

    private Map<String, String> initialJNDIProperties;

    @Override
    protected void doInit() throws JBIException {
        // Save initialJNDIProperties
        saveJNDIProperties();

        this.ejbPool = new HashMap<Provides, Object>();
        this.suClassLoaders = new HashMap<String, ClassLoader>();
        this.xmlEnginePool = new HashMap<Provides, XMLStream>();
        this.dataBindingServicePool = new HashMap<Provides, Service>();
        this.setServiceUnitListener(new EjbServiceUnitListener(this));
    }

    public Object getEjb(Provides provides) {
        return ejbPool.get(provides);
    }


    public void addEjb(Provides provides, Object ejb) {
        ejbPool.put(provides, ejb);
    }

    public void removeEjb(Provides provides) {
        ejbPool.remove(provides);
    }

    public XMLStream getXMLEngine(Provides provides) {
        return this.xmlEnginePool.get(provides);
    }

    public void addXMLEngine(Provides provides, XMLStream xmlEngine) {
        this.xmlEnginePool.put(provides, xmlEngine);
    }

    public void removeXMLEngine(Provides provides) {
        this.xmlEnginePool.remove(provides);
    }

    public void setDataBindingServicePool(Map<Provides, Service> jbiServicePool) {
        this.dataBindingServicePool = jbiServicePool;
    }

    public Map<Provides, Service> getDataBindingServicePool() {
        return dataBindingServicePool;
    }

    public ClassLoader getSuClassLoader(Provides provides) {
        String suName = ((BindingComponentServiceUnitManager) this
                .getServiceUnitManager()).getSUDataHandlerForService(provides)
                .getName();
        return suClassLoaders.get(suName);
    }

    public void addSuClassLoader(String suName, ClassLoader classLoader) {
        suClassLoaders.put(suName, classLoader);
    }

    public void removeClassLoader(String suName) {
        suClassLoaders.remove(suName);
    }

    private void saveJNDIProperties() {
        this.initialJNDIProperties = new HashMap<String, String>();

        this.initialJNDIProperties.put(Context.INITIAL_CONTEXT_FACTORY, System
                .getProperty(Context.INITIAL_CONTEXT_FACTORY));
        this.initialJNDIProperties.put(Context.URL_PKG_PREFIXES, System
                .getProperty(Context.URL_PKG_PREFIXES));
        this.initialJNDIProperties.put(Context.PROVIDER_URL, System
                .getProperty(Context.PROVIDER_URL));

        this.initialJNDIProperties.put(Context.SECURITY_AUTHENTICATION, System
                .getProperty(Context.SECURITY_AUTHENTICATION));
        this.initialJNDIProperties.put(Context.SECURITY_CREDENTIALS, System
                .getProperty(Context.SECURITY_CREDENTIALS));
        this.initialJNDIProperties.put(Context.SECURITY_PRINCIPAL, System
                .getProperty(Context.SECURITY_PRINCIPAL));
    }

    public void resetJNDIProperties() {
        Properties systemProperties = System.getProperties();

        for (String key : this.initialJNDIProperties.keySet()) {
            String value = this.initialJNDIProperties.get(key);
            if (value != null) {
                System.setProperty(key, value);
            } else {
                if (systemProperties.getProperty(key) != null) {
                    systemProperties.remove(key);
                }
            }
        }
    }

    public Context createInitialContext(ConfigurationExtensions extensions) throws PEtALSCDKException {
        Properties environment = new Properties();

        // Retrieve the provider URl from the SU configuration
        String providerUrl = extensions.get(ProvidesConstants.URL);
        if ((providerUrl == null) || (providerUrl.equals(""))) {
            throw new PEtALSCDKException(
                    "You must define a provider url in the SU configuration");
        }
        environment.put(Context.PROVIDER_URL, providerUrl);

        // Retrieve the initial context factory from the SU configuration
        String initialContextFactory = extensions.get(ProvidesConstants.INITIAL_CONTEXT_FACTORY);
        if ((initialContextFactory == null) || (initialContextFactory.equals(""))) {
            throw new PEtALSCDKException(
                    "You must define an inital context factory in the SU configuration");
        }
        environment.put(Context.INITIAL_CONTEXT_FACTORY, initialContextFactory);

        // Retrieve the security principal from the SU configuration
        String securityPrincipal = extensions.get(ProvidesConstants.SECURITY_PRINCIPAL);

        // Retrieve the security credentials from the SU configuration
        String securityCredentials = extensions.get(ProvidesConstants.SECURITY_CREDENTIALS);

        if (!((securityPrincipal == null) || (securityPrincipal.equals("")))) {
            if (!((securityCredentials == null) || (securityCredentials.equals("")))) {
                environment.put(Context.SECURITY_PRINCIPAL, securityPrincipal);
                environment.put(Context.SECURITY_CREDENTIALS,
                        securityCredentials);
            }
        }

        // Creates the JNDI context
        return createJNDIContext(environment);
    }

    public Context createJNDIContext(Properties properties) throws PEtALSCDKException {
        Context ctx = null;
        ClassLoader cl = this.getClass().getClassLoader();
        Class<?> clazz = null;

        try {
            clazz = cl.loadClass(properties.getProperty(Context.INITIAL_CONTEXT_FACTORY));
        } catch (ClassNotFoundException e) {
            try {
                clazz = Thread.currentThread().getContextClassLoader().loadClass(
                        properties.getProperty(Context.INITIAL_CONTEXT_FACTORY));
            } catch (ClassNotFoundException e1) {
                throw new PEtALSCDKException(
                        "Can't load initial context factory class :" + e.getMessage(), e);
            }
        }
        InitialContextFactory icf;
        try {
            icf = (InitialContextFactory) clazz.newInstance();
            ctx = icf.getInitialContext(properties);
        } catch (Exception e) {
            throw new PEtALSCDKException(
                    "Can't create the JNDI context :" + e.getMessage(), e);
        }
        return ctx;
    }

    private Class<?> loadEjbHomeClass(ConfigurationExtensions extensions)
            throws PEtALSCDKException {
        Class<?> clazz = null;
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        try {
            clazz = cl.loadClass(extensions
                    .get(ProvidesConstants.EJB_HOME_INTERFACE_PROPERTY));
        } catch (ClassNotFoundException e) {
            throw new PEtALSCDKException("Can't load ejb home interface class",
                    e);
        }
        return clazz;
    }

    public Object retrieveObject(Context initialContext,
            ConfigurationExtensions extensions) throws PEtALSCDKException {
        Object ejb = null;

        String ejbVersion = extensions
                .get(ProvidesConstants.EJB_VERSION_PROPERTY);
        if (ejbVersion != null) {
            String ejbJndiName = extensions
                    .get(ProvidesConstants.EJB_JNDI_NAME_PROPERTY);

            if (ejbVersion.equals(EjbVersion.EJB_VERSION_3_0) || ejbVersion.equals(EjbVersion.EJB_VERSION_3_1)) {
                try {
                    ejb = initialContext.lookup(ejbJndiName);
                } catch (NamingException e) {
                    throw new PEtALSCDKException(
                            "Can't retrieve ejb 3 for name :" + ejbJndiName, e);
                }
            } else if (ejbVersion.equals(EjbVersion.EJB_VERSION_2_1) || ejbVersion.equals(EjbVersion.EJB_VERSION_2_0)) {
                try {
                    Object ref = initialContext.lookup(ejbJndiName);
                    Class<?> homeInterface = loadEjbHomeClass(extensions);
                    Object home = PortableRemoteObject.narrow(ref,
                            homeInterface);
                    ejb = createEjb2x(home);
                } catch (NamingException e) {
                    throw new PEtALSCDKException(
                            "Can't retrieve ejb 2.x home interface for name :"
                                    + ejbJndiName, e);
                }
            } else {
                throw new PEtALSCDKException(
                        "You must specify a supported EJB version in your SU extensions. "
                                + "Supported EJB version are "
                                + EjbVersion.EJB_VERSION_2_0 + ", "
                                + EjbVersion.EJB_VERSION_2_1 + ", "
                                + EjbVersion.EJB_VERSION_3_0 + " and "
                                + EjbVersion.EJB_VERSION_3_1);
            }
        } else {
            throw new PEtALSCDKException("You must specify an EJB version ("
                    + ProvidesConstants.EJB_VERSION_PROPERTY
                    + " property) in your SU extensions");
        }
        return ejb;
    }

    private Object createEjb2x(Object home) throws PEtALSCDKException {
        Object ejb2 = null;
        // Search the method to invoke
        Method methodToInvoke;
        try {
            methodToInvoke = home.getClass()
                    .getMethod("create", new Class[] {});
        } catch (NoSuchMethodException e) {
            throw new PEtALSCDKException(
                    "Can't create an ejb stub from the given EJB home interface ("
                            + home.getClass()
                            + "). Your home interface must define create() method without arguments.");
        }

        try {
            ejb2 = methodToInvoke.invoke(home, new Object[] {});
        } catch (SecurityException e) {
            throw new PEtALSCDKException(
                    "Error invoking create() method on EJB home : security error",
                    e);
        } catch (IllegalArgumentException e) {
            throw new PEtALSCDKException(
                    "Error invoking create() method on EJB, bad method arguments",
                    e);
        } catch (Exception e) {
            throw new PEtALSCDKException(
                    "Error invoking create() method on EJB", e);
        }
        return ejb2;
    }

    
}
