/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ServiceManager.java ofabre$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.bc.ejb.listener;

import java.io.ByteArrayInputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.naming.Context;
import javax.security.auth.Subject;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.ow2.petals.bc.ejb.constants.ProvidesConstants;
import org.ow2.petals.bc.ejb.EjbBC;
import org.ow2.petals.bc.ejb.model.Params;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.api.message.Exchange;
//import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.XMLUtil;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import org.ow2.petals.databinding.jaxb.service.Service;

/**
 *
 * @author ofabre & ccasadei - eBM Websourcing
 */
public class ServiceManager {

    ///////////////////////////////////////////////////////////////////////////
    // Fields
    /**
     * 
     */
    private AbstractJBIListener listener;

    /**
     * 
     */
    private EjbBC component;

    ///////////////////////////////////////////////////////////////////////////
    // Constructors & init
    /**
     * 
     * @param listener
     */
    public ServiceManager(AbstractJBIListener listener) {
        super();
        this.component = (EjbBC) listener.getComponent();
        this.listener = listener;
    }

    ///////////////////////////////////////////////////////////////////////////
    // Methods
    /**
     * 
     * @param exchange
     * @throws org.ow2.petals.component.framework.api.exception.PEtALSCDKException
     */
    public void onJBIMessage(Exchange exchange) throws PEtALSCDKException {
        listener.getLogger().log(Level.FINE, "Try to call an EJB");

        // Operation to call
        QName operation = exchange.getOperation();

        // Get in message content
        Document requestDocument;
        try {
            requestDocument = exchange.getInMessageContentAsDocument();
        } catch (MessagingException e) {
            throw new PEtALSCDKException(
                    "Can't retrieve in message content for exchange : "
                            + exchange.getExchangeId(), e);
        }

        // Retrieve SU extensions
        ConfigurationExtensions extensions = listener.getExtensions();

        // Retrieve Su classloader
        ClassLoader suClassloader = this.component.getSuClassLoader(listener
                .getProvides());

        // Change context classloader to component classloader
        ClassLoader oldCL = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(suClassloader);

        LoginContext loginContext = null;
        try {
            // Retrieve the EJB Object for this exchange
            Object ejb = this.component.getEjb(listener.getProvides());

            // Check if a security subject exists in MessageExchange to login
            Subject subject = exchange.getInMessage().getSecuritySubject();
            if (subject != null) {
                loginContext = login(extensions, subject);
            }

            // Call ejb
             Document responseDocument;
             if (extensions.get(ProvidesConstants.MARSHALLING_ENGINE).equals("xstream"))
                 responseDocument = this.callEjb(ejb, extensions,
                         operation, requestDocument);
             else
                 responseDocument = this.callEjbUsingJAXB(ejb, extensions,
                 operation, requestDocument);

            if (responseDocument != null
                    && (exchange.isInOptionalOutPattern() || exchange
                            .isInOutPattern())) {
                try {
                    exchange.setOutMessageContent(responseDocument);
                } catch (MessagingException e) {
                    throw new PEtALSCDKException(
                            "Can't set out message content for exchange : "
                                    + exchange.getExchangeId(), e);
                }
            }
        } finally {
            // reset old classloader
            Thread.currentThread().setContextClassLoader(oldCL);

            // Logout
            if (loginContext != null) {
                logout(loginContext);
            }
        }
    }
    
    /**
     * 
     * @param ejb
     * @param extensions
     * @param operation
     * @param requestDocument
     * @return
     * @throws PEtALSCDKException
     */
    private Document callEjbUsingJAXB(Object ejb, ConfigurationExtensions extensions,
            QName operation, Document requestDocument) throws PEtALSCDKException {
        
        // Get the JbiService object to Marshall/Unmarshall the message content
        Source messageRequest = new DOMSource(requestDocument);
        Service jbiService = this.component.getDataBindingServicePool().get(listener.getProvides());
        
        Document messageResult = null;
        Object[] methodParamValue = null;
        Object methodResult = null;
        
        // Create the method parameters from the incoming message
        try {
            methodParamValue = jbiService.getServiceUnmarshaller().unmarshallMethodParam(
                    jbiService.getOperationByName(operation.getLocalPart()), messageRequest);
        } catch (Exception e) {
            throw new PEtALSCDKException("DataBinding Error : Can't create method parameters from the incoming message ", e);
        } 
        
        // Call the EJB method
        if (methodParamValue != null) {
            try {
                Method methodToInvoke = ejb.getClass().getMethod(operation.getLocalPart(),
                        jbiService.getOperationByName(operation.getLocalPart()).getParameterTypes());
                System.err.println(methodToInvoke.getName());
                methodResult = methodToInvoke.invoke(ejb, methodParamValue);
            } catch (Exception e) {
                throw new PEtALSCDKException(" Error while the call of the EJB method ", e);
            }
        }
        
        // Marhsall the method result to create the output message content
        if (methodResult != null) {
            try {
                StreamSource messageResponse = (StreamSource) jbiService.getServiceMarshaller().marshallMethodResult(
                        jbiService.getOperationByName(operation.getLocalPart()), new Object[]{methodResult});
                DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
                DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
                messageResult = documentBuilder.parse(messageResponse.getInputStream());
            } catch (Exception e) {
                throw new PEtALSCDKException(" DataBinding Error : can't create the message response" +
                        " from the method result  ", e);
            } 
        }
        return messageResult;
    }

    /**
     * 
     * @param extensions
     * @param subject
     * @return
     * @throws org.ow2.petals.component.framework.api.exception.PEtALSCDKException
     */
    private LoginContext login(ConfigurationExtensions extensions,
            Subject subject) throws PEtALSCDKException {

        LoginContext loginContext = null;

        // Retrieve security configuration name
        String securityConf = extensions.get(ProvidesConstants.SECURITY_NAME);

        if (securityConf != null) {
            try {
                loginContext = new LoginContext(securityConf, subject);
                loginContext.login();
            } catch (LoginException e) {
                throw new PEtALSCDKException("Login failed", e);
            }
        }

        return loginContext;
    }

    /**
     *
     * @param loginContext
     * @throws org.ow2.petals.component.framework.api.exception.PEtALSCDKException
     */
    private void logout(LoginContext loginContext) throws PEtALSCDKException {
        try {
            loginContext.logout();
        } catch (LoginException e) {
            throw new PEtALSCDKException("Failed to logout", e);
        }
    }

    private Document callEjb(Object ejb, ConfigurationExtensions extensions,
            QName operation, Document requestDocument)
            throws PEtALSCDKException {
        // Create an bean representation
        Params params = retrieveEjbMethodModel(requestDocument);
        Object result = null;

        // Search the method to invoke
        Method methodToInvoke;
        try {
            methodToInvoke = ejb.getClass().getMethod(operation.getLocalPart(),
                    params.getTypes());
        } catch (NoSuchMethodException e) {
            throw new PEtALSCDKException("Can't find method '"
                    + operation.getLocalPart() + "' in EJB interface : "
                    + ejb.getClass().getName(), e);
        }

        try {
            result = methodToInvoke.invoke(ejb, params.getParams());
        } catch (SecurityException e) {
            throw new PEtALSCDKException("Error invoking EJB : security error",
                    e);
        } catch (IllegalArgumentException e) {
            throw new PEtALSCDKException(
                    "Error invoking EJB, bad method arguments", e);
        } catch (IllegalAccessException e) {
            throw new PEtALSCDKException("Error invoking EJB", e);
        } catch (InvocationTargetException e) {
            try {
                // Try to rebind ejb in case of application server connection
                // loss
                Context initialContext = this.component
                        .createInitialContext(extensions);
                Object newEjb = this.component.retrieveObject(initialContext,
                        extensions);
                this.component.addEjb(listener.getProvides(), newEjb);
                // Try a second invocation
                result = methodToInvoke.invoke(newEjb, params.getParams());
            } catch (Exception e1) {
                throw new PEtALSCDKException("Error invoking EJB", e);
            } finally {
                // reset jndi system properties to avoid conflicts with other
                // component
                this.component.resetJNDIProperties();
            }
        }

        Document response = null;
        if (result != null) {
            response = createResponseDocument(result);
        }

        return response;
    }

    private Document createResponseDocument(Object result)
            throws PEtALSCDKException {
        Document document = null;
        try {
            String stringResult = this.component.getXMLEngine(
                    listener.getProvides()).toXML(result);
            DocumentBuilderFactory factory = DocumentBuilderFactory
                    .newInstance();
            DocumentBuilder documentBuilder = factory.newDocumentBuilder();
            document = documentBuilder.parse(new ByteArrayInputStream(
                    stringResult.getBytes()));
        } catch (Exception e) {
            throw new PEtALSCDKException("Error getting field value", e);
        }
        return document;
    }

    private Params retrieveEjbMethodModel(Document requestDocument)
            throws PEtALSCDKException {
        Params params = new Params();
        Element element = requestDocument.getDocumentElement();
        NodeList children = element.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            Node node = children.item(i);
            try {
                if (!(node.getNodeType() == Node.TEXT_NODE)) {
                    Object object = this.component.getXMLEngine(
                            listener.getProvides()).fromXML(
                            XMLUtil.createStringFromDOMNode(node));
                    params.addParam(object);
                    params.addParamType(object.getClass());
                }
            } catch (Throwable e) {
                throw new PEtALSCDKException("Can't read input param : "
                        + node.getLocalName(), e);
            }

        }
        return params;
    }
}