/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: EjbServiceUnitListener.java ccasadei$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.bc.ejb.su;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;


import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.Context;

import org.ow2.petals.bc.ejb.EjbBC;
import org.ow2.petals.bc.ejb.constants.MarshallingEngine;
import org.ow2.petals.bc.ejb.constants.ProvidesConstants;
import org.ow2.petals.bc.ejb.xmlengine.XMLStream;
import org.ow2.petals.bc.ejb.xmlengine.XStreamEngine;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.su.ServiceUnitListener;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.jbidescriptor.generated.Services;
import org.ow2.petals.component.framework.util.ClassLoaderUtil;
import org.ow2.petals.databinding.jaxb.DataBindingException;
import org.ow2.petals.databinding.jaxb.ServiceFactory;
import org.ow2.petals.databinding.jaxb.service.Service;
import org.ow2.petals.databinding.jaxb.tools.JaxbClassGenerator;
//import org.ow2.petals.scorware.runtime.wsdl2jbi.ServiceFactory;
//import org.ow2.petals.scorware.runtime.wsdl2jbi.service.Operation;
//import org.ow2.petals.scorware.runtime.wsdl2jbi.service.Service;
//import org.ow2.petals.scorware.runtime.wsdl2jbi.util.JaxbClassGenerator;

public class EjbServiceUnitListener implements ServiceUnitListener {

    private EjbBC component;

    public EjbServiceUnitListener(EjbBC ejbBC) {
        super();
        this.component = ejbBC;
    }

    public void onSUDeployed(final String serviceUnitName,
            final String suRootPath, final Jbi descriptor)
    throws PEtALSCDKException {
        ClassLoader suClassLoader = this.createSUClassLoader(suRootPath);
        this.component.addSuClassLoader(serviceUnitName, suClassLoader);

        //Changes the context classloader to the component classloader
        ClassLoader oldClassLoader = Thread.currentThread()
        .getContextClassLoader();
        Thread.currentThread().setContextClassLoader(suClassLoader);

        try {
            for (Provides provides : this.getProvidesList(descriptor)) {
                this.registerEjb(provides, suRootPath);
                ConfigurationExtensions extensions = null;
                if (provides.getAny().size() > 0) {
                    extensions = new ConfigurationExtensions(provides.getAny());
                } else {
                    throw new PEtALSCDKException(
                    "Error : The compenent attributes configuration are not defined ");
                }
                if (extensions.get(ProvidesConstants.MARSHALLING_ENGINE).equals("xstream"))
                    this.createXMLStream(provides, suClassLoader);
                else 
                    this.createJAXBDatabinding(provides, suClassLoader, suRootPath);
            }
        } catch (PEtALSCDKException e) {
            throw new PEtALSCDKException(
                    "Error creating a connection pool for SU: " + suRootPath, e);
        } finally {
            //Reset the JNDI system to avoid conflicts with other components
            this.component.resetJNDIProperties();

            //Reseting the ClassLoader
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
    }

    private void createJAXBDatabinding(Provides provides, ClassLoader suClassLoader, String suRootPath) throws PEtALSCDKException {
      
        // Save the system properties
        Properties properties = (Properties) System.getProperties().clone();
        try {
            // Get the WSDL file name for the provide service
            String wsdlFileName = provides.getWsdl();
            // Get the service Name to generate the directory name
            String serviceName = provides.getServiceName().getLocalPart();
            // Get the component class path
            URL[] seClassPath = ClassLoaderUtil.getUrls(this.component.getContext().getInstallRoot());
            List<String> classPath = new ArrayList<String>();
            for (URL url : seClassPath) {
                classPath.add(url.getPath());
            }
            // Call JAXB to generate the DataBinding classes for the EJB interface
            String wsdlFilePath = suRootPath + File.separator + wsdlFileName;
            String jaxbBaseDir = suRootPath + serviceName + "-jaxb";
            new File(jaxbBaseDir).mkdir();
            JaxbClassGenerator.generateJaxbClass(wsdlFilePath, jaxbBaseDir, serviceName);
            // Compile the generated JAXB classes
            JaxbClassGenerator.compileJaxbClass(jaxbBaseDir, serviceName, classPath);
            // Create the Service object to Marshall/Unmarshall the request/response messages
            // parameters and Marshall method result
            String jaxbClassesUrl = "file://" + jaxbBaseDir + "/target/classes/";
            ServiceFactory serviceFactory = ServiceFactory.getInstance();
            Service service = serviceFactory.crateService(wsdlFilePath, jaxbClassesUrl, suClassLoader, this.component.getEjb(provides).getClass());
            component.getDataBindingServicePool().put(provides, service);
        } catch (DataBindingException ex) {
            throw new PEtALSCDKException("Databinding error");
        }
        
        finally {
            // Restore the system properties
            System.setProperties(properties);
        }

    }



    public void onSUStarted(String arg0) throws PEtALSCDKException {
        // TODO Auto-generated method stub
        // Nothing to do right here
    }

    public void onSUStopped(String arg0) throws PEtALSCDKException {
        // TODO Auto-generated method stub
        // Nothing to do right here
    }

    public void onSUUndeployed(final String serviceUnitName,
            final String suRootPath, final Jbi descriptor)
    throws PEtALSCDKException {
        for (Provides provides : getProvidesList(descriptor)) {
            component.removeEjb(provides);
            component.removeXMLEngine(provides);
            component.getDataBindingServicePool().remove(provides);
        }
        component.removeClassLoader(serviceUnitName);
    }

    private void registerEjb(Provides provides, String suRootPath)
    throws PEtALSCDKException {
        ConfigurationExtensions extensions = null;

        if (provides.getAny().size() > 0) {
            extensions = new ConfigurationExtensions(provides.getAny());
        } else {
            throw new PEtALSCDKException(
                    "Error while creating the XML stream : SU configuration is empty");
        }

        if (extensions != null) {
            // Create an initial context according to given JNDI parameters
            Context initialContext = component.createInitialContext(extensions);

            // Retrieve the EJB Object matching the given name
            Object ejb = component.retrieveObject(initialContext, extensions);

            // Register ejb
            component.addEjb(provides, ejb);
        }
    }

    private void createXMLStream(Provides provides, ClassLoader suClassLoader)
    throws PEtALSCDKException {
        ConfigurationExtensions extensions = null;

        if (provides.getAny().size() > 0) {
            extensions = new ConfigurationExtensions(provides.getAny());
        } else {
            throw new PEtALSCDKException(
            "Error while creating the XML stream : SU configuration is empty");
        }

        if (extensions != null) {
            XMLStream xmlstream = null;

            //Selecting the appropriate XML Engine using the SU configuration
            String suDefinedMarshallingEngine = extensions
            .get(ProvidesConstants.MARSHALLING_ENGINE);
            if (suDefinedMarshallingEngine.equals(MarshallingEngine.XSTREAM)) {
                xmlstream = new XStreamEngine(extensions, suClassLoader);
            } else {
                throw new PEtALSCDKException(
                        "Error while creating the XML stream : "
                        + suDefinedMarshallingEngine
                        + " didn't exists.");
            }

            this.component.addXMLEngine(provides, xmlstream);
        }
    }

    private ClassLoader createSUClassLoader(String suRootPath) {
        return ClassLoaderUtil.createClassLoader(suRootPath, this.getClass()
                .getClassLoader());
    }

    private List<Provides> getProvidesList(Jbi descriptor) {
        List<Provides> providesList = new ArrayList<Provides>();
        Services services = descriptor.getServices();
        if (services != null) {
            providesList = services.getProvides();
        }
        return providesList;
    }


}