/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XMLEngine.java ccasadei$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.bc.ejb.xmlengine;

import com.thoughtworks.xstream.XStream;
import java.io.ByteArrayInputStream;
import java.util.Map;
import java.util.StringTokenizer;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.ow2.petals.bc.ejb.EjbBC;
import org.ow2.petals.bc.ejb.constants.ProvidesConstants;
import org.ow2.petals.bc.ejb.model.Params;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.w3c.dom.Document;

/**
 * 
 * @author ccasadei - eBM WebSourcing
 */
public class XStreamEngine extends XMLStream {

	private ConfigurationExtensions extensions;

	private XStream xstream;

	public XStreamEngine(ConfigurationExtensions extensions,
			ClassLoader suClassLoader) throws PEtALSCDKException {
		this.extensions = extensions;

		if (extensions != null) {
			xstream = new XStream();

			// Add aliases for xml marshalling
			for (Map.Entry<String, String> param : extensions.entrySet()) {
				// If we're parsing an alias
				if (param.getKey().equals(
						ProvidesConstants.XSTREAM_ALIAS.toString())) {
					String[] parsedRawAlias = null;

					// Parse rawAlias
					parsedRawAlias = this.parseXStreamAlias(param.getValue());

					Class clazz = this.loadAliasClass(suClassLoader,
							parsedRawAlias[1]);
					this.xstream.alias(parsedRawAlias[0], clazz);
				}
			}
		}
	}

	/**
	 * Splits the alias string to get the alias and its value.
	 * 
	 * @param rawXStreamAlias
	 * @return
	 * @throws
	 *            org.ow2.petals.component.framework.api.exception.PEtALSCDKException
	 */
	private String[] parseXStreamAlias(String rawXStreamAlias)
			throws PEtALSCDKException {
		String[] result = null;

		// Splitting the raw XStream alias
		result = rawXStreamAlias.split("\\|");

		// Cleaning the generated aliases & values
		for (int i = 0; i < result.length; i++) {
			result[i] = result[i].trim();
		}

		if (result.length > 2) {
			throw new PEtALSCDKException(
					"Error while creating the XML stream : wrong XStream alias format");
		}

		return result;
	}

	private Class loadAliasClass(ClassLoader suClassLoader, String value)
			throws PEtALSCDKException {
		try {
			return suClassLoader.loadClass(value);
		} catch (ClassNotFoundException e) {
			throw new PEtALSCDKException("Can't load xml stream alias class : "
					+ value, e);
		}
	}

	public String toXML(Object o) {
		return this.xstream.toXML(o);
	}

	public Object fromXML(String xmlstream) {
		return this.xstream.fromXML(xmlstream);
	}
}
